##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Standard configuration errors
"""

import traceback

__all__ = [
    'ConfigurationError',
]


class ConfigurationError(Exception):
    """There was an error in a configuration
    """

    # A list of strings or things that can be converted to strings,
    # added by append_details as we walk back up the call/include stack.
    # We will print them in reverse order so that the most recent detail is
    # last.
    _details = ()

    def add_details(self, info):
        if isinstance(info, BaseException):
            info = traceback.format_exception_only(type(info), info)
            # Trim trailing newline
            info[-1] = info[-1].rstrip()
            self._details += tuple(info)
        else:
            self._details += (info,)
        return self

    def _with_details(self, opening, detail_formatter):
        lines = ['    ' + detail_formatter(detail) for detail in self._details]
        lines.append(opening)
        lines.reverse()
        return '\n'.join(lines)

    def __str__(self):
        s = super(ConfigurationError, self).__str__()
        return self._with_details(s, str)

    def __repr__(self):
        s = super(ConfigurationError, self).__repr__()
        return self._with_details(s, repr)


class ConfigurationWrapperError(ConfigurationError):

    USE_INFO_REPR = False

    def __init__(self, info, exception):
        super(ConfigurationWrapperError, self).__init__(
            repr(info) if self.USE_INFO_REPR else info)
        self.add_details(exception)

        # This stuff is undocumented and not used but we store
        # for backwards compatibility
        self.info = info
        self.etype = type(exception)
        self.evalue = exception



##############################################################################
#
# Copyright (c) 20!2 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test zope.configuration.__init__
"""
import unittest


class Test_namespace(unittest.TestCase):

    def _callFUT(self, *args, **kw):
        from zope.configuration import namespace
        return namespace(*args, **kw)

    def test_empty(self):
        self.assertEqual(self._callFUT(''),
                         'http://namespaces.zope.org/')

    def test_non_empty(self):
        self.assertEqual(self._callFUT('test'),
                         'http://namespaces.zope.org/test')



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Enable "Making specific directives condition" section of narrative docs.
"""
from zope.interface import Interface
from zope.schema import Id


class IRegister(Interface):
    """Trivial sample registry."""

    id = Id(
        title=u"Identifier",
        description=u"Some identifier that can be checked.",
        required=True,
    )


registry = []


def register(context, id):
    context.action(discriminator=('Register', id),
                   callable=registry.append,
                   args=(id,)
                   )



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Utilities for the 'nested directive' section in the narrative docs.
"""

from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implementer
from zope.schema import NativeStringLine
from zope.schema import Id
from zope.schema import Int
from zope.schema import Text
from zope.schema import TextLine
from zope.configuration.config import GroupingContextDecorator
from zope.configuration.fields import Bool


schema_registry = {}


class ISchemaInfo(Interface):
    """Parameter schema for the schema directive
    """

    name = TextLine(
        title=u"The schema name",
        description=u"This is a descriptive name for the schema.",
    )

    id = Id(title=u"The unique id for the schema")


class ISchema(Interface):
    """Interface that distinguishes the schema directive
    """

    fields = Attribute("Dictionary of field definitions")


@implementer(ISchema)
class Schema(GroupingContextDecorator):
    """Handle schema directives
    """

    def __init__(self, context, name, id):
        self.context, self.name, self.id = context, name, id
        self.fields = {}

    def after(self):
        schema = Interface.__class__(
            self.name,
            (Interface, ),
            self.fields
        )
        schema.__doc__ = self.info.text.strip()
        self.action(
            discriminator=('schema', self.id),
            callable=schema_registry.__setitem__,
            args=(self.id, schema),
        )


class IFieldInfo(Interface):

    name = NativeStringLine(
        title=u"The field name",
    )

    title = TextLine(
        title=u"Title",
        description=u"A short summary or label",
        default=u"",
        required=False,
    )

    required = Bool(
        title=u"Required",
        description=u"Determines whether a value is required.",
        default=True)

    readonly = Bool(
        title=u"Read Only",
        description=u"Can the value be modified?",
        required=False,
        default=False)


class ITextInfo(IFieldInfo):

    min_length = Int(
        title=u"Minimum length",
        description=(
            u"Value after whitespace processing cannot have less than "
            u"min_length characters. If min_length is None, there is "
            u"no minimum."
        ),
        required=False,
        min=0,  # needs to be a positive number
        default=0)

    max_length = Int(
        title=u"Maximum length",
        description=(
            u"Value after whitespace processing cannot have greater "
            u"or equal than max_length characters. If max_length is "
            u"None, there is no maximum."
        ),
        required=False,
        min=0,  # needs to be a positive number
        default=None)


def field(context, constructor, name, **kw):

    # Compute the field
    field = constructor(description=context.info.text.strip(), **kw)

    # Save it in the schema's field dictionary
    schema = context.context
    if name in schema.fields:
        raise ValueError("Duplicate field", name)
    schema.fields[name] = field


def textField(context, **kw):
    field(context, Text, **kw)


class IIntInfo(IFieldInfo):

    min = Int(
        title=u"Start of the range",
        required=False,
        default=None
    )

    max = Int(
        title=u"End of the range (excluding the value itself)",
        required=False,
        default=None
    )


def intField(context, **kw):
    field(context, Int, **kw)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test directives
"""
from zope.interface import Interface
from zope.interface import implementer
from zope.schema import NativeStringLine
from zope.schema import Text

from zope.configuration.config import GroupingContextDecorator
from zope.configuration.interfaces import IConfigurationContext
from zope.configuration.fields import GlobalObject


class F(object):
    def __repr__(self):
        return 'f'

    def __call__(self, *a, **k):
        raise NotImplementedError


f = F()


class ISimple(Interface):

    a = Text()
    b = Text(required=False)
    c = NativeStringLine()


def simple(context, a=None, c=None, b=u"xxx"):
    return [(('simple', a, b, c), f, (a, b, c))]


def newsimple(context, a, c, b=u"xxx"):
    context.action(('newsimple', a, b, c), f, (a, b, c))


class IPackaged(Interface):

    package = GlobalObject()


class IPackagedContext(IPackaged, IConfigurationContext):
    pass


@implementer(IPackagedContext)
class Packaged(GroupingContextDecorator):
    pass


class IFactory(Interface):

    factory = GlobalObject()


def factory(context, factory):
    context.action(('factory', 1, 2), factory)


class Complex(object):

    def __init__(self, context, a, c, b=u"xxx"):
        self.a, self.b, self.c = a, b, c
        context.action("Complex.__init__")

    def factory(self, context, factory):
        return [(('Complex.factory', 1, 2), factory, (self.a, ))]

    def __call__(self):
        return [(('Complex', 1, 2), f, (self.b, self.c))]


class Ik(Interface):
    for_ = NativeStringLine()
    class_ = NativeStringLine()
    x = NativeStringLine()


def k(context, for_, class_, x):
    context.action(('k', for_), f, (for_, class_, x))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Provide configuration object name resolution.

.. note:: This module is no longer used by `zope.configuration` and
   may be deprecated soon. Its functions are not documented.
"""

import os
from types import ModuleType

__all__ = [
    'resolve',
    'getNormalizedName',
    'path',
]


def resolve(name, package='zopeproducts', _silly=('__doc__',), _globals={}):
    name = name.strip()

    if name.startswith('.'):
        name = package + name

    if name.endswith('.') or name.endswith('+'):
        name = name[:-1]
        repeat = True
    else:
        repeat = False

    names = name.split('.')
    last = names[-1]
    mod = '.'.join(names[:-1])

    if not mod:
        return __import__(name, _globals, _globals, _silly)

    while True:
        m = __import__(mod, _globals, _globals, _silly)
        try:
            a = getattr(m, last)
        except AttributeError:
            if not repeat:
                return __import__(name, _globals, _globals, _silly)

        else:
            if not repeat or (not isinstance(a, ModuleType)):
                return a
        mod += '.' + last


def getNormalizedName(name, package):
    name = name.strip()
    if name.startswith('.'):
        name = package + name

    if name.endswith('.') or name.endswith('+'):
        name = name[:-1]
        repeat = True
    else:
        repeat = False
    name = name.split(".")
    while len(name) > 1 and name[-1] == name[-2]:
        name.pop()
        repeat = 1
    name = ".".join(name)
    if repeat:
        name += "+"
    return name


def path(file='', package='zopeproducts', _silly=('__doc__',), _globals={}):
    # XXX WTF? why not look for abspath before importing?
    try:
        package = __import__(package, _globals, _globals, _silly)
    except ImportError:
        norm = os.path.normpath(file)
        if file and os.path.abspath(norm) == norm:
            # The package didn't matter
            return norm
        raise

    path = os.path.dirname(package.__file__)

    if file:
        path = os.path.join(path, file)

    return path



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests for for zope.configuration.docutils
"""
import unittest


class Test_wrap(unittest.TestCase):

    def _callFUT(self, *args, **kw):
        from zope.configuration.docutils import wrap
        return wrap(*args, **kw)

    def test_empty(self):
        self.assertEqual(self._callFUT(''), '\n\n')

    def test_only_whitespace(self):
        self.assertEqual(self._callFUT(' \t\n\r'), '\n\n')

    def test_single_paragraphs(self):
        self.assertEqual(
            self._callFUT('abcde fghij klmno pqrst uvwxy', 10, 3),
            '   abcde\n   fghij\n   klmno\n   pqrst\n   uvwxy\n\n')

    def test_multiple_paragraphs(self):
        self.assertEqual(
            self._callFUT('abcde fghij klmno\n\npqrst uvwxy', 10, 3),
            '   abcde\n   fghij\n   klmno\n\n   pqrst\n   uvwxy\n\n')


class Test_makeDocStructures(unittest.TestCase):

    def _callFUT(self, *args, **kw):
        from zope.configuration.docutils import makeDocStructures
        return makeDocStructures(*args, **kw)

    def _makeContext(self):
        class _Context(object):
            def __init__(self):
                self._docRegistry = []
        return _Context()

    def test_empty(self):
        context = self._makeContext()
        namespaces, subdirs = self._callFUT(context)
        self.assertEqual(len(namespaces), 0)
        self.assertEqual(len(subdirs), 0)

    def test_wo_parents(self):
        from zope.interface import Interface

        class ISchema(Interface):
            pass

        class IUsedIn(Interface):
            pass
        NS = 'http://namespace.example.com/main'
        NS2 = 'http://namespace.example.com/other'

        def _one():
            raise AssertionError("should not be called")

        def _two():
            raise AssertionError("should not be called")

        def _three():
            raise AssertionError("should not be called")
        context = self._makeContext()
        context._docRegistry.append(
            ((NS, 'one'), ISchema, IUsedIn, _one, 'ONE', None))
        context._docRegistry.append(
            ((NS2, 'two'), ISchema, IUsedIn, _two, 'TWO', None))
        context._docRegistry.append(
            ((NS, 'three'), ISchema, IUsedIn, _three, 'THREE', None))
        namespaces, subdirs = self._callFUT(context)
        self.assertEqual(len(namespaces), 2)
        self.assertEqual(namespaces[NS], {'one': (ISchema, _one, 'ONE'),
                                          'three': (ISchema, _three, 'THREE')})
        self.assertEqual(namespaces[NS2], {'two': (ISchema, _two, 'TWO')})
        self.assertEqual(len(subdirs), 0)

    def test_w_parents(self):
        from zope.interface import Interface

        class ISchema(Interface):
            pass

        class IUsedIn(Interface):
            pass
        PNS = 'http://namespace.example.com/parent'
        NS = 'http://namespace.example.com/main'
        NS2 = 'http://namespace.example.com/other'

        def _one():
            raise AssertionError("should not be called")

        def _two():
            raise AssertionError("should not be called")

        def _three():
            raise AssertionError("should not be called")

        class Parent(object):
            namespace = PNS
            name = 'parent'
        parent1 = Parent()
        parent2 = Parent()
        parent2.name = 'parent2'
        context = self._makeContext()
        context._docRegistry.append(
            ((NS, 'one'), ISchema, IUsedIn, _one, 'ONE', parent1))
        context._docRegistry.append(
            ((NS2, 'two'), ISchema, IUsedIn, _two, 'TWO', parent2))
        context._docRegistry.append(
            ((NS, 'three'), ISchema, IUsedIn, _three, 'THREE', parent1))
        namespaces, subdirs = self._callFUT(context)
        self.assertEqual(len(namespaces), 0)
        self.assertEqual(len(subdirs), 2)
        self.assertEqual(subdirs[(PNS, 'parent')],
                         [(NS, 'one', ISchema, _one, 'ONE'),
                          (NS, 'three', ISchema, _three, 'THREE')])
        self.assertEqual(subdirs[(PNS, 'parent2')],
                         [(NS2, 'two', ISchema, _two, 'TWO')])



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import sys

PY3 = sys.version_info[0] >= 3

if PY3:  # pragma: PY3

    import builtins

    string_types = (str,)
    text_type = str

    # borrowed from 'six'
    def reraise(tp, value, tb=None):
        try:  # pragma: no cover
            if value is None:
                value = tp
            if value.__traceback__ is not tb:
                raise value.with_traceback(tb)
            raise value
        finally:
            value = None
            tb = None

else:  # pragma: PY2

    import __builtin__ as builtins  # noqa: F401 imported but unused

    text_type = unicode  # noqa: F821 undefined name
    string_types = (basestring,)  # noqa: F821 undefined name

    # borrowed from 'six'
    exec("""\
def reraise(tp, value, tb=None):
    try:
        raise tp, value, tb
    finally:
        tb = None
""")


class implementer_if_needed(object):
    # Helper to make sure we don't redundantly implement
    # interfaces already inherited. Doing so tends to produce
    # problems with the C3 order. In this package, we could easily
    # statically determine to elide the relevant interfaces, but
    # this is a defense against changes in parent classes and lessens
    # the testing burden.
    def __init__(self, *ifaces):
        self._ifaces = ifaces

    def __call__(self, cls):
        from zope.interface import implementedBy
        from zope.interface import implementer

        ifaces_needed = []
        implemented = implementedBy(cls)
        ifaces_needed = [
            iface
            for iface in self._ifaces
            if not implemented.isOrExtends(iface)
        ]
        return implementer(*ifaces_needed)(cls)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Configuration processor
"""
from keyword import iskeyword
import operator
import os.path
import sys

from zope.interface.adapter import AdapterRegistry
from zope.interface import Interface
from zope.interface import implementer
from zope.interface import providedBy
from zope.schema import TextLine
from zope.schema import URI
from zope.schema import ValidationError

from zope.configuration.exceptions import ConfigurationError
from zope.configuration.exceptions import ConfigurationWrapperError
from zope.configuration.interfaces import IConfigurationContext
from zope.configuration.interfaces import IGroupingContext
from zope.configuration.fields import GlobalInterface
from zope.configuration.fields import GlobalObject
from zope.configuration.fields import PathProcessor
from zope.configuration._compat import builtins
from zope.configuration._compat import reraise
from zope.configuration._compat import string_types
from zope.configuration._compat import text_type
from zope.configuration._compat import implementer_if_needed

__all__ = [
    'ConfigurationContext',
    'ConfigurationAdapterRegistry',
    'ConfigurationMachine',
    'IStackItem',
    'SimpleStackItem',
    'RootStackItem',
    'GroupingStackItem',
    'ComplexStackItem',
    'GroupingContextDecorator',
    'DirectiveSchema',
    'IDirectivesInfo',
    'IDirectivesContext',
    'DirectivesHandler',
    'IDirectiveInfo',
    'IFullInfo',
    'IStandaloneDirectiveInfo',
    'defineSimpleDirective',
    'defineGroupingDirective',
    'IComplexDirectiveContext',
    'ComplexDirectiveDefinition',
    'subdirective',
    'IProvidesDirectiveInfo',
    'provides',
    'toargs',
    'expand_action',
    'resolveConflicts',
    'ConfigurationConflictError',
]

zopens = 'http://namespaces.zope.org/zope'
metans = 'http://namespaces.zope.org/meta'
testns = 'http://namespaces.zope.org/test'


class ConfigurationContext(object):
    """
    Mix-in for implementing.
    :class:`zope.configuration.interfaces.IConfigurationContext`.

    Note that this class itself does not actually declare that it
    implements that interface; the subclass must do that. In addition,
    subclasses must provide a ``package`` attribute and a ``basepath``
    attribute. If the base path is not None, relative paths are
    converted to absolute paths using the the base path. If the
    package is not none, relative imports are performed relative to
    the package.

    In general, the basepath and package attributes should be
    consistent. When a package is provided, the base path should be
    set to the path of the package directory.

    Subclasses also provide an ``actions`` attribute, which is a list
    of actions, an ``includepath`` attribute, and an ``info``
    attribute.

    The include path is appended to each action and is used when
    resolving conflicts among actions. Normally, only the a
    ConfigurationMachine provides the actions attribute. Decorators
    simply use the actions of the context they decorate. The
    ``includepath`` attribute is a tuple of names. Each name is
    typically the name of an included configuration file.

    The ``info`` attribute contains descriptive information helpful
    when reporting errors. If not set, it defaults to an empty string.

    The actions attribute is a sequence of dictionaries where each
    dictionary has the following keys:

        - ``discriminator``, a value that identifies the action. Two
          actions that have the same (non None) discriminator
          conflict.

        - ``callable``, an object that is called to execute the
          action,

        - ``args``, positional arguments for the action

        - ``kw``, keyword arguments for the action

        - ``includepath``, a tuple of include file names (defaults to
          ())

        - ``info``, an object that has descriptive information about
          the action (defaults to '')
    """

    # pylint:disable=no-member

    def __init__(self):
        super(ConfigurationContext, self).__init__()
        self._seen_files = set()
        self._features = set()

    def resolve(self, dottedname):
        """
        Resolve a dotted name to an object.

        Examples:

             >>> from zope.configuration.config import ConfigurationContext
             >>> c = ConfigurationContext()
             >>> import zope, zope.interface
             >>> c.resolve('zope') is zope
             True
             >>> c.resolve('zope.interface') is zope.interface
             True
             >>> c.resolve('zope.configuration.eek') #doctest: +NORMALIZE_WHITESPACE
             Traceback (most recent call last):
             ...
             ConfigurationError:
             ImportError: Module zope.configuration has no global eek

             >>> c.resolve('.config.ConfigurationContext')
             Traceback (most recent call last):
             ...
             AttributeError: 'ConfigurationContext' object has no attribute 'package'
             >>> import zope.configuration
             >>> c.package = zope.configuration
             >>> c.resolve('.') is zope.configuration
             True
             >>> c.resolve('.config.ConfigurationContext') is ConfigurationContext
             True
             >>> c.resolve('..interface') is zope.interface
             True
             >>> c.resolve('str') is str
             True
        """  # noqa: E501 line too long
        name = dottedname.strip()

        if not name:
            raise ValueError("The given name is blank")

        if name == '.':
            return self.package

        names = name.split('.')

        if not names[-1]:
            raise ValueError(
                "Trailing dots are no longer supported in dotted names")

        if len(names) == 1:
            # Check for built-in objects
            marker = object()
            obj = getattr(builtins, names[0], marker)
            if obj is not marker:
                return obj

        if not names[0]:
            # Got a relative name. Convert it to abs using package info
            if self.package is None:
                raise ConfigurationError(
                    "Can't use leading dots in dotted names, "
                    "no package has been set.")
            pnames = self.package.__name__.split(".")
            pnames.append('')
            while names and not names[0]:
                names.pop(0)
                try:
                    pnames.pop()
                except IndexError:
                    raise ConfigurationError("Invalid global name", name)
            names[0:0] = pnames

        # Now we should have an absolute dotted name

        # Split off object name:
        oname, mname = names[-1], '.'.join(names[:-1])

        # Import the module
        if not mname:
            # Just got a single name. Must me a module
            mname = oname
            oname = ''

        try:
            # Without a fromlist, this returns the package rather than the
            # module if the name contains a dot.  Using a fromlist requires
            # star imports to work, which may not be true if there are
            # unicode items in __all__ due to unicode_literals on Python 2.
            # Getting the module from sys.modules instead avoids both
            # problems.
            __import__(mname)
            mod = sys.modules[mname]
        except ImportError as v:
            if sys.exc_info()[2].tb_next is not None:
                # ImportError was caused deeper
                raise
            raise ConfigurationError(
                "ImportError: Couldn't import %s, %s" % (mname, v))

        if not oname:
            # see not mname case above
            return mod

        try:
            obj = getattr(mod, oname)
            return obj
        except AttributeError:
            # No such name, maybe it's a module that we still need to import
            try:
                moname = mname + '.' + oname
                __import__(moname)
                return sys.modules[moname]
            except ImportError:
                if sys.exc_info()[2].tb_next is not None:
                    # ImportError was caused deeper
                    raise
                raise ConfigurationError(
                    "ImportError: Module %s has no global %s" % (mname, oname))

    def path(self, filename):
        """
        Compute package-relative paths.

        Examples:

             >>> import os
             >>> from zope.configuration.config import ConfigurationContext
             >>> c = ConfigurationContext()
             >>> c.path("/x/y/z") == os.path.normpath("/x/y/z")
             True
             >>> c.path("y/z")
             Traceback (most recent call last):
             ...
             AttributeError: 'ConfigurationContext' object has no attribute 'package'
             >>> import zope.configuration
             >>> c.package = zope.configuration
             >>> import os
             >>> d = os.path.dirname(zope.configuration.__file__)
             >>> c.path("y/z") == d + os.path.normpath("/y/z")
             True
             >>> c.path("y/./z") == d + os.path.normpath("/y/z")
             True
             >>> c.path("y/../z") == d + os.path.normpath("/z")
             True
        """  # noqa: E501 line too long
        filename, needs_processing = PathProcessor.expand(filename)

        if not needs_processing:
            return filename

        # Got a relative path, combine with base path.
        # If we have no basepath, compute the base path from the package
        # path.
        basepath = getattr(self, 'basepath', '')

        if not basepath:
            if self.package is None:
                basepath = os.getcwd()
            else:
                if hasattr(self.package, '__path__'):
                    basepath = self.package.__path__[0]
                else:
                    basepath = os.path.dirname(self.package.__file__)
                basepath = os.path.abspath(os.path.normpath(basepath))
            self.basepath = basepath

        return os.path.normpath(os.path.join(basepath, filename))

    def checkDuplicate(self, filename):
        """
        Check for duplicate imports of the same file.

        Raises an exception if this file had been processed before.
        This is better than an unlimited number of conflict errors.

        Examples:

             >>> from zope.configuration.config import ConfigurationContext
             >>> from zope.configuration.config import ConfigurationError
             >>> c = ConfigurationContext()
             >>> c.checkDuplicate('/foo.zcml')
             >>> try:
             ...     c.checkDuplicate('/foo.zcml')
             ... except ConfigurationError as e:
             ...     # On Linux the exact msg has /foo, on Windows \\foo.
             ...     str(e).endswith("foo.zcml' included more than once")
             True

        You may use different ways to refer to the same file:

             >>> import zope.configuration
             >>> c.package = zope.configuration
             >>> import os
             >>> d = os.path.dirname(zope.configuration.__file__)
             >>> c.checkDuplicate('bar.zcml')
             >>> try:
             ...   c.checkDuplicate(d + os.path.normpath('/bar.zcml'))
             ... except ConfigurationError as e:
             ...   str(e).endswith("bar.zcml' included more than once")
             ...
             True

        """
        path = self.path(filename)
        if path in self._seen_files:
            raise ConfigurationError('%r included more than once' % path)
        self._seen_files.add(path)

    def processFile(self, filename):
        """
        Check whether a file needs to be processed.

        Return True if processing is needed and False otherwise. If
        the file needs to be processed, it will be marked as
        processed, assuming that the caller will procces the file if
        it needs to be procssed.

        Examples:

             >>> from zope.configuration.config import ConfigurationContext
             >>> c = ConfigurationContext()
             >>> c.processFile('/foo.zcml')
             True
             >>> c.processFile('/foo.zcml')
             False

        You may use different ways to refer to the same file:

             >>> import zope.configuration
             >>> c.package = zope.configuration
             >>> import os
             >>> d = os.path.dirname(zope.configuration.__file__)
             >>> c.processFile('bar.zcml')
             True
             >>> c.processFile(os.path.join(d, 'bar.zcml'))
             False
        """
        path = self.path(filename)
        if path in self._seen_files:
            return False
        self._seen_files.add(path)
        return True

    def action(self, discriminator, callable=None, args=(), kw=None, order=0,
               includepath=None, info=None, **extra):
        """
        Add an action with the given discriminator, callable and
        arguments.

        For testing purposes, the callable and arguments may be
        omitted. In that case, a default noop callable is used.

        The discriminator must be given, but it can be None, to
        indicate that the action never conflicts.


        Examples:

             >>> from zope.configuration.config import ConfigurationContext
             >>> c = ConfigurationContext()

        Normally, the context gets actions from subclasses. We'll provide
        an actions attribute ourselves:

             >>> c.actions = []

        We'll use a test callable that has a convenient string representation

             >>> from zope.configuration.tests.directives import f
             >>> c.action(1, f, (1, ), {'x': 1})
             >>> from pprint import PrettyPrinter
             >>> pprint = PrettyPrinter(width=60).pprint
             >>> pprint(c.actions)
             [{'args': (1,),
               'callable': f,
               'discriminator': 1,
               'includepath': (),
               'info': '',
               'kw': {'x': 1},
               'order': 0}]

             >>> c.action(None)
             >>> pprint(c.actions)
             [{'args': (1,),
               'callable': f,
               'discriminator': 1,
               'includepath': (),
               'info': '',
               'kw': {'x': 1},
               'order': 0},
              {'args': (),
               'callable': None,
               'discriminator': None,
               'includepath': (),
               'info': '',
               'kw': {},
               'order': 0}]

        Now set the include path and info:

             >>> c.includepath = ('foo.zcml',)
             >>> c.info = "?"
             >>> c.action(None)
             >>> pprint(c.actions[-1])
             {'args': (),
              'callable': None,
              'discriminator': None,
              'includepath': ('foo.zcml',),
              'info': '?',
              'kw': {},
              'order': 0}

        We can add an order argument to crudely control the order
        of execution:

             >>> c.action(None, order=99999)
             >>> pprint(c.actions[-1])
             {'args': (),
              'callable': None,
              'discriminator': None,
              'includepath': ('foo.zcml',),
              'info': '?',
              'kw': {},
              'order': 99999}

        We can also pass an includepath argument, which will be used as the the
        includepath for the action.  (if includepath is None, self.includepath
        will be used):

             >>> c.action(None, includepath=('abc',))
             >>> pprint(c.actions[-1])
             {'args': (),
              'callable': None,
              'discriminator': None,
              'includepath': ('abc',),
              'info': '?',
              'kw': {},
              'order': 0}

        We can also pass an info argument, which will be used as the the
        source line info for the action.  (if info is None, self.info will be
        used):

             >>> c.action(None, info='abc')
             >>> pprint(c.actions[-1])
             {'args': (),
              'callable': None,
              'discriminator': None,
              'includepath': ('foo.zcml',),
              'info': 'abc',
              'kw': {},
              'order': 0}

        """
        if kw is None:
            kw = {}

        action = extra

        if info is None:
            info = getattr(self, 'info', '')

        if includepath is None:
            includepath = getattr(self, 'includepath', ())

        action.update(
            dict(
                discriminator=discriminator,
                callable=callable,
                args=args,
                kw=kw,
                includepath=includepath,
                info=info,
                order=order,
            )
        )

        self.actions.append(action)

    def hasFeature(self, feature):
        """
        Check whether a named feature has been provided.

        Initially no features are provided.

        Examples:

            >>> from zope.configuration.config import ConfigurationContext
            >>> c = ConfigurationContext()
            >>> c.hasFeature('onlinehelp')
            False

        You can declare that a feature is provided

            >>> c.provideFeature('onlinehelp')

        and it becomes available

            >>> c.hasFeature('onlinehelp')
            True
        """
        return feature in self._features

    def provideFeature(self, feature):
        """
        Declare that a named feature has been provided.

        See :meth:`hasFeature` for examples.
        """
        self._features.add(feature)


class ConfigurationAdapterRegistry(object):
    """
    Simple adapter registry that manages directives as adapters.

    Examples:

        >>> from zope.configuration.interfaces import IConfigurationContext
        >>> from zope.configuration.config import ConfigurationAdapterRegistry
        >>> from zope.configuration.config import ConfigurationMachine
        >>> r = ConfigurationAdapterRegistry()
        >>> c = ConfigurationMachine()
        >>> r.factory(c, ('http://www.zope.com', 'xxx'))
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Unknown directive', 'http://www.zope.com', 'xxx')
        >>> def f():
        ...     pass

        >>> r.register(
        ...     IConfigurationContext, ('http://www.zope.com', 'xxx'), f)
        >>> r.factory(c, ('http://www.zope.com', 'xxx')) is f
        True
        >>> r.factory(c, ('http://www.zope.com', 'yyy')) is f
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Unknown directive', 'http://www.zope.com', 'yyy')
        >>> r.register(IConfigurationContext, 'yyy', f)
        >>> r.factory(c, ('http://www.zope.com', 'yyy')) is f
        True

    Test the documentation feature:

        >>> from zope.configuration.config import IFullInfo
        >>> r._docRegistry
        []
        >>> r.document(('ns', 'dir'), IFullInfo, IConfigurationContext, None,
        ...            'inf', None)
        >>> r._docRegistry[0][0] == ('ns', 'dir')
        True
        >>> r._docRegistry[0][1] is IFullInfo
        True
        >>> r._docRegistry[0][2] is IConfigurationContext
        True
        >>> r._docRegistry[0][3] is None
        True
        >>> r._docRegistry[0][4] == 'inf'
        True
        >>> r._docRegistry[0][5] is None
        True
        >>> r.document('all-dir', None, None, None, None)
        >>> r._docRegistry[1][0]
        ('', 'all-dir')
    """

    def __init__(self):
        super(ConfigurationAdapterRegistry, self).__init__()
        self._registry = {}
        # Stores tuples of form:
        #   (namespace, name), schema, usedIn, info, parent
        self._docRegistry = []

    def register(self, interface, name, factory):
        r = self._registry.get(name)
        if r is None:
            r = AdapterRegistry()
            self._registry[name] = r

        r.register([interface], Interface, '', factory)

    def document(self, name, schema, usedIn, handler, info, parent=None):
        if isinstance(name, string_types):
            name = ('', name)
        self._docRegistry.append((name, schema, usedIn, handler, info, parent))

    def factory(self, context, name):
        r = self._registry.get(name)
        if r is None:
            # Try namespace-independent name
            ns, n = name
            r = self._registry.get(n)
            if r is None:
                raise ConfigurationError("Unknown directive", ns, n)

        f = r.lookup1(providedBy(context), Interface)
        if f is None:
            raise ConfigurationError(
                "The directive %s cannot be used in this context" % (name, ))
        return f


@implementer(IConfigurationContext)
class ConfigurationMachine(ConfigurationAdapterRegistry, ConfigurationContext):
    """
    Configuration machine.

    Example:

      >>> from zope.configuration.config import ConfigurationMachine
      >>> machine = ConfigurationMachine()
      >>> ns = "http://www.zope.org/testing"

    Register a directive:

      >>> from zope.configuration.config import metans
      >>> machine((metans, "directive"),
      ...         namespace=ns, name="simple",
      ...         schema="zope.configuration.tests.directives.ISimple",
      ...         handler="zope.configuration.tests.directives.simple")

    and try it out:

      >>> machine((ns, "simple"), a=u"aa", c=u"cc")
      >>> from pprint import PrettyPrinter
      >>> pprint = PrettyPrinter(width=60).pprint
      >>> pprint(machine.actions)
      [{'args': ('aa', 'xxx', 'cc'),
        'callable': f,
        'discriminator': ('simple', 'aa', 'xxx', 'cc'),
        'includepath': (),
        'info': None,
        'kw': {},
        'order': 0}]
    """
    package = None
    basepath = None
    includepath = ()
    info = ''

    #: These `Exception` subclasses are allowed to be raised from
    #: `execute_actions` without being re-wrapped into a
    #: `~.ConfigurationError`. (`BaseException` and other
    #: `~.ConfigurationError` instances are never wrapped.)
    #:
    #: Users of instances of this class may modify this before calling
    #: `execute_actions` if they need to propagate specific exceptions.
    #:
    #: .. versionadded:: 4.2.0
    pass_through_exceptions = ()

    def __init__(self):
        super(ConfigurationMachine, self).__init__()
        self.actions = []
        self.stack = [RootStackItem(self)]
        self.i18n_strings = {}
        _bootstrap(self)

    def begin(self, __name, __data=None, __info=None, **kw):
        if __data:
            if kw:
                raise TypeError("Can't provide a mapping object and keyword "
                                "arguments")
        else:
            __data = kw
        self.stack.append(self.stack[-1].contained(__name, __data, __info))

    def end(self):
        self.stack.pop().finish()

    def __call__(self, __name, __info=None, **__kw):
        self.begin(__name, __kw, __info)
        self.end()

    def getInfo(self):
        return self.stack[-1].context.info

    def setInfo(self, info):
        self.stack[-1].context.info = info

    def execute_actions(self, clear=True, testing=False):
        """
        Execute the configuration actions.

        This calls the action callables after resolving conflicts.

        For example:

            >>> from zope.configuration.config import ConfigurationMachine
            >>> output = []
            >>> def f(*a, **k):
            ...    output.append(('f', a, k))
            >>> context = ConfigurationMachine()
            >>> context.actions = [
            ...   (1, f, (1,)),
            ...   (1, f, (11,), {}, ('x', )),
            ...   (2, f, (2,)),
            ...   ]
            >>> context.execute_actions()
            >>> output
            [('f', (1,), {}), ('f', (2,), {})]

        If the action raises an error, we convert it to a
        `~.ConfigurationError`.

            >>> output = []
            >>> def bad():
            ...    bad.xxx
            >>> context.actions = [
            ...   (1, f, (1,)),
            ...   (1, f, (11,), {}, ('x', )),
            ...   (2, f, (2,)),
            ...   (3, bad, (), {}, (), 'oops')
            ...   ]

            >>> context.execute_actions()
            Traceback (most recent call last):
            ...
            zope.configuration.config.ConfigurationExecutionError: oops
                AttributeError: 'function' object has no attribute 'xxx'

        Note that actions executed before the error still have an effect:

            >>> output
            [('f', (1,), {}), ('f', (2,), {})]

        If the exception was already a `~.ConfigurationError`, it is raised
        as-is with the action's ``info`` added.

            >>> def bad():
            ...     raise ConfigurationError("I'm bad")
            >>> context.actions = [
            ...   (1, f, (1,)),
            ...   (1, f, (11,), {}, ('x', )),
            ...   (2, f, (2,)),
            ...   (3, bad, (), {}, (), 'oops')
            ...   ]
            >>> context.execute_actions()
            Traceback (most recent call last):
            ...
            zope.configuration.exceptions.ConfigurationError: I'm bad
                oops

        """
        pass_through_exceptions = self.pass_through_exceptions
        if testing:
            pass_through_exceptions = BaseException
        try:
            for action in resolveConflicts(self.actions):
                callable = action['callable']
                if callable is None:
                    continue
                args = action['args']
                kw = action['kw']
                info = action['info']
                try:
                    callable(*args, **kw)
                except ConfigurationError as ex:
                    ex.add_details(info)
                    raise
                except pass_through_exceptions:
                    raise
                except Exception:
                    # Wrap it up and raise.
                    reraise(
                        ConfigurationExecutionError(info, sys.exc_info()[1]),
                        None, sys.exc_info()[2])
        finally:
            if clear:
                del self.actions[:]


class ConfigurationExecutionError(ConfigurationWrapperError):
    """
    An error occurred during execution of a configuration action
    """

##############################################################################
# Stack items


class IStackItem(Interface):
    """
    Configuration machine stack items

    Stack items are created when a directive is being processed.

    A stack item is created for each directive use.
    """

    def contained(name, data, info):
        """Begin processing a contained directive

        The data are a dictionary of attribute names mapped to unicode
        strings.

        The info argument is an object that can be converted to a
        string and that contains information about the directive.

        The begin method returns the next item to be placed on the stack.
        """

    def finish():
        """Finish processing a directive
        """


@implementer(IStackItem)
class SimpleStackItem(object):
    """
    Simple stack item

    A simple stack item can't have anything added after it.  It can
    only be removed.  It is used for simple directives and
    subdirectives, which can't contain other directives.

    It also defers any computation until the end of the directive
    has been reached.
    """
    # XXX why this *argdata hack instead of schema, data?

    def __init__(self, context, handler, info, *argdata):
        newcontext = GroupingContextDecorator(context)
        newcontext.info = info
        self.context = newcontext
        self.handler = handler
        self.argdata = argdata

    def contained(self, name, data, info):
        raise ConfigurationError("Invalid directive %s" % str(name))

    def finish(self):
        # We're going to use the context that was passed to us, which wasn't
        # created for the directive.  We want to set it's info to the one
        # passed to us while we make the call, so we'll save the old one
        # and restore it.
        context = self.context
        args = toargs(context, *self.argdata)
        actions = self.handler(context, **args)
        if actions:
            # we allow the handler to return nothing
            for action in actions:
                if not isinstance(action, dict):
                    action = expand_action(*action)  # b/c
                context.action(**action)


@implementer(IStackItem)
class RootStackItem(object):
    """
    A root stack item.
    """

    def __init__(self, context):
        self.context = context

    def contained(self, name, data, info):
        """Handle a contained directive

        We have to compute a new stack item by getting a named adapter
        for the current context object.
        """
        factory = self.context.factory(self.context, name)
        if factory is None:
            raise ConfigurationError("Invalid directive", name)
        adapter = factory(self.context, data, info)
        return adapter

    def finish(self):
        pass


@implementer_if_needed(IStackItem)
class GroupingStackItem(RootStackItem):
    """
    Stack item for a grouping directive

    A grouping stack item is in the stack when a grouping directive is
    being processed. Grouping directives group other directives.
    Often, they just manage common data, but they may also take
    actions, either before or after contained directives are executed.

    A grouping stack item is created with a grouping directive
    definition, a configuration context, and directive data.

    To see how this works, let's look at an example:

    We need a context. We'll just use a configuration machine

        >>> from zope.configuration.config import GroupingStackItem
        >>> from zope.configuration.config import ConfigurationMachine
        >>> context = ConfigurationMachine()

    We need a callable to use in configuration actions. We'll use a
    convenient one from the tests:

        >>> from zope.configuration.tests.directives import f

    We need a handler for the grouping directive. This is a class that
    implements a context decorator. The decorator must also provide
    ``before`` and ``after`` methods that are called before and after
    any contained directives are processed. We'll typically subclass
    ``GroupingContextDecorator``, which provides context decoration,
    and default ``before`` and ``after`` methods.

        >>> from zope.configuration.config import GroupingContextDecorator
        >>> class SampleGrouping(GroupingContextDecorator):
        ...    def before(self):
        ...       self.action(('before', self.x, self.y), f)
        ...    def after(self):
        ...       self.action(('after'), f)

    We'll use our decorator to decorate our initial context, providing
    keyword arguments x and y:

        >>> dec = SampleGrouping(context, x=1, y=2)

    Note that the keyword arguments are made attributes of the
    decorator.

    Now we'll create the stack item.

        >>> item = GroupingStackItem(dec)

    We still haven't called the before action yet, which we can verify
    by looking at the context actions:

        >>> context.actions
        []

    Subdirectives will get looked up as adapters of the context.

    We'll create a simple handler:

        >>> def simple(context, data, info):
        ...     context.action(("simple", context.x, context.y, data), f)
        ...     return info

    and register it with the context:

        >>> from zope.configuration.interfaces import IConfigurationContext
        >>> from zope.configuration.config import testns
        >>> context.register(IConfigurationContext, (testns, 'simple'), simple)

    This handler isn't really a propert handler, because it doesn't
    return a new context. It will do for this example.

    Now we'll call the contained method on the stack item:

        >>> item.contained((testns, 'simple'), {'z': 'zope'}, "someinfo")
        'someinfo'

    We can verify thet the simple method was called by looking at the
    context actions. Note that the before method was called before
    handling the contained directive.

        >>> from pprint import PrettyPrinter
        >>> pprint = PrettyPrinter(width=60).pprint

        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': ('before', 1, 2),
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': ('simple', 1, 2, {'z': 'zope'}),
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0}]

    Finally, we call finish, which calls the decorator after method:

        >>> item.finish()

        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': ('before', 1, 2),
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': ('simple', 1, 2, {'z': 'zope'}),
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': 'after',
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0}]

    If there were no nested directives:

        >>> context = ConfigurationMachine()
        >>> dec = SampleGrouping(context, x=1, y=2)
        >>> item = GroupingStackItem(dec)
        >>> item.finish()

    Then before will be when we call finish:

        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': ('before', 1, 2),
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': 'after',
          'includepath': (),
          'info': '',
          'kw': {},
          'order': 0}]
    """

    def __init__(self, context):
        super(GroupingStackItem, self).__init__(context)

    def __callBefore(self):
        actions = self.context.before()
        if actions:
            for action in actions:
                if not isinstance(action, dict):
                    action = expand_action(*action)
                self.context.action(**action)
        self.__callBefore = noop

    def contained(self, name, data, info):
        self.__callBefore()
        return RootStackItem.contained(self, name, data, info)

    def finish(self):
        self.__callBefore()
        actions = self.context.after()
        if actions:
            for action in actions:
                if not isinstance(action, dict):
                    action = expand_action(*action)
                self.context.action(**action)


def noop():
    pass


@implementer(IStackItem)
class ComplexStackItem(object):
    """
    Complex stack item

    A complex stack item is in the stack when a complex directive is
    being processed. It only allows subdirectives to be used.

    A complex stack item is created with a complex directive
    definition (IComplexDirectiveContext), a configuration context,
    and directive data.

    To see how this works, let's look at an example:

    We need a context. We'll just use a configuration machine

        >>> from zope.configuration.config import ConfigurationMachine
        >>> context = ConfigurationMachine()

    We need a callable to use in configuration actions. We'll use a
    convenient one from the tests:

        >>> from zope.configuration.tests.directives import f

    We need a handler for the complex directive. This is a class with
    a method for each subdirective:

        >>> class Handler(object):
        ...   def __init__(self, context, x, y):
        ...      self.context, self.x, self.y = context, x, y
        ...      context.action('init', f)
        ...   def sub(self, context, a, b):
        ...      context.action(('sub', a, b), f)
        ...   def __call__(self):
        ...      self.context.action(('call', self.x, self.y), f)

    We need a complex directive definition:

        >>> from zope.interface import Interface
        >>> from zope.schema import TextLine
        >>> from zope.configuration.config import ComplexDirectiveDefinition
        >>> class Ixy(Interface):
        ...    x = TextLine()
        ...    y = TextLine()
        >>> definition = ComplexDirectiveDefinition(
        ...        context, name="test", schema=Ixy,
        ...        handler=Handler)
        >>> class Iab(Interface):
        ...    a = TextLine()
        ...    b = TextLine()
        >>> definition['sub'] = Iab, ''

    OK, now that we have the context, handler and definition, we're
    ready to use a stack item.

        >>> from zope.configuration.config import ComplexStackItem
        >>> item = ComplexStackItem(
        ...     definition, context, {'x': u'xv', 'y': u'yv'}, 'foo')

    When we created the definition, the handler (factory) was called.

        >>> from pprint import PrettyPrinter
        >>> pprint = PrettyPrinter(width=60).pprint
        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': 'init',
          'includepath': (),
          'info': 'foo',
          'kw': {},
          'order': 0}]

    If a subdirective is provided, the ``contained`` method of the
    stack item is called. It will lookup the subdirective schema and
    call the corresponding method on the handler instance:

        >>> simple = item.contained(('somenamespace', 'sub'),
        ...                         {'a': u'av', 'b': u'bv'}, 'baz')
        >>> simple.finish()

    Note that the name passed to ``contained`` is a 2-part name,
    consisting of a namespace and a name within the namespace.

        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': 'init',
          'includepath': (),
          'info': 'foo',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': ('sub', 'av', 'bv'),
          'includepath': (),
          'info': 'baz',
          'kw': {},
          'order': 0}]

    The new stack item returned by contained is one that doesn't allow
    any more subdirectives,

    When all of the subdirectives have been provided, the ``finish``
    method is called:

        >>> item.finish()

    The stack item will call the handler if it is callable.

        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': 'init',
          'includepath': (),
          'info': 'foo',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': ('sub', 'av', 'bv'),
          'includepath': (),
          'info': 'baz',
          'kw': {},
          'order': 0},
         {'args': (),
          'callable': f,
          'discriminator': ('call', 'xv', 'yv'),
          'includepath': (),
          'info': 'foo',
          'kw': {},
          'order': 0}]
    """

    def __init__(self, meta, context, data, info):
        newcontext = GroupingContextDecorator(context)
        newcontext.info = info
        self.context = newcontext
        self.meta = meta

        # Call the handler contructor
        args = toargs(newcontext, meta.schema, data)
        self.handler = self.meta.handler(newcontext, **args)

    def contained(self, name, data, info):
        """Handle a subdirective
        """
        # Look up the subdirective meta data on our meta object
        ns, name = name
        schema = self.meta.get(name)
        if schema is None:
            raise ConfigurationError("Invalid directive", name)
        schema = schema[0]  # strip off info
        handler = getattr(self.handler, name)
        return SimpleStackItem(self.context, handler, info, schema, data)

    def finish(self):
        # when we're done, we call the handler, which might return more actions
        # Need to save and restore old info
        # XXX why not just use callable()?
        try:
            actions = self.handler()
        except AttributeError as v:
            if v.args[0] == '__call__':
                return  # noncallable
            raise
        except TypeError:
            return  # non callable

        if actions:
            # we allow the handler to return nothing
            for action in actions:
                if not isinstance(action, dict):
                    action = expand_action(*action)
                self.context.action(**action)


##############################################################################
# Helper classes

@implementer(IConfigurationContext, IGroupingContext)
class GroupingContextDecorator(ConfigurationContext):
    """Helper mix-in class for building grouping directives

    See the discussion (and test) in GroupingStackItem.
    """

    def __init__(self, context, **kw):
        self.context = context
        for name, v in kw.items():
            setattr(self, name, v)

    def __getattr__(self, name):
        v = getattr(self.context, name)
        # cache result in self
        setattr(self, name, v)
        return v

    def before(self):
        pass

    def after(self):
        pass

##############################################################################
# Directive-definition


class DirectiveSchema(GlobalInterface):
    """A field that contains a global variable value that must be a schema
    """


class IDirectivesInfo(Interface):
    """Schema for the ``directives`` directive
    """

    namespace = URI(
        title=u"Namespace",
        description=(
            u"The namespace in which directives' names "
            u"will be defined"),
    )


class IDirectivesContext(IDirectivesInfo, IConfigurationContext):
    pass


@implementer(IDirectivesContext)
class DirectivesHandler(GroupingContextDecorator):
    """Handler for the directives directive

    This is just a grouping directive that adds a namespace attribute
    to the normal directive context.

    """


class IDirectiveInfo(Interface):
    """Information common to all directive definitions.
    """

    name = TextLine(
        title=u"Directive name",
        description=u"The name of the directive being defined",
    )

    schema = DirectiveSchema(
        title=u"Directive handler",
        description=u"The dotted name of the directive handler",
    )


class IFullInfo(IDirectiveInfo):
    """Information that all top-level directives (not subdirectives)
    have.
    """

    handler = GlobalObject(
        title=u"Directive handler",
        description=u"The dotted name of the directive handler",
    )

    usedIn = GlobalInterface(
        title=u"The directive types the directive can be used in",
        description=(
            u"The interface of the directives that can contain "
            u"the directive"
        ),
        default=IConfigurationContext,
    )


class IStandaloneDirectiveInfo(IDirectivesInfo, IFullInfo):
    """Info for full directives defined outside a directives directives
    """


def defineSimpleDirective(context, name, schema, handler,
                          namespace='', usedIn=IConfigurationContext):
    """
    Define a simple directive

    Define and register a factory that invokes the simple directive
    and returns a new stack item, which is always the same simple
    stack item.

    If the namespace is '*', the directive is registered for all
    namespaces.

    Example:

        >>> from zope.configuration.config import ConfigurationMachine
        >>> context = ConfigurationMachine()
        >>> from zope.interface import Interface
        >>> from zope.schema import TextLine
        >>> from zope.configuration.tests.directives import f
        >>> class Ixy(Interface):
        ...    x = TextLine()
        ...    y = TextLine()
        >>> def s(context, x, y):
        ...    context.action(('s', x, y), f)

        >>> from zope.configuration.config import defineSimpleDirective
        >>> defineSimpleDirective(context, 's', Ixy, s, testns)

        >>> context((testns, "s"), x=u"vx", y=u"vy")
        >>> from pprint import PrettyPrinter
        >>> pprint = PrettyPrinter(width=60).pprint
        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': ('s', 'vx', 'vy'),
          'includepath': (),
          'info': None,
          'kw': {},
          'order': 0}]

        >>> context(('http://www.zope.com/t1', "s"), x=u"vx", y=u"vy")
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Unknown directive', 'http://www.zope.com/t1', 's')

        >>> context = ConfigurationMachine()
        >>> defineSimpleDirective(context, 's', Ixy, s, "*")

        >>> context(('http://www.zope.com/t1', "s"), x=u"vx", y=u"vy")
        >>> pprint(context.actions)
        [{'args': (),
          'callable': f,
          'discriminator': ('s', 'vx', 'vy'),
          'includepath': (),
          'info': None,
          'kw': {},
          'order': 0}]
    """  # noqa: E501 line too long
    namespace = namespace or context.namespace
    if namespace != '*':
        name = namespace, name

    def factory(context, data, info):
        return SimpleStackItem(context, handler, info, schema, data)
    factory.schema = schema

    context.register(usedIn, name, factory)
    context.document(name, schema, usedIn, handler, context.info)


def defineGroupingDirective(context, name, schema, handler,
                            namespace='', usedIn=IConfigurationContext):
    """
    Define a grouping directive

    Define and register a factory that sets up a grouping directive.

    If the namespace is '*', the directive is registered for all
    namespaces.

    Example:

        >>> from zope.configuration.config import ConfigurationMachine
        >>> context = ConfigurationMachine()
        >>> from zope.interface import Interface
        >>> from zope.schema import TextLine
        >>> class Ixy(Interface):
        ...    x = TextLine()
        ...    y = TextLine()

    We won't bother creating a special grouping directive class. We'll
    just use :class:`GroupingContextDecorator`, which simply sets up a
    grouping context that has extra attributes defined by a schema:

        >>> from zope.configuration.config import defineGroupingDirective
        >>> from zope.configuration.config import GroupingContextDecorator
        >>> defineGroupingDirective(context, 'g', Ixy,
        ...                         GroupingContextDecorator, testns)

        >>> context.begin((testns, "g"), x=u"vx", y=u"vy")
        >>> context.stack[-1].context.x
        'vx'
        >>> context.stack[-1].context.y
        'vy'

        >>> context(('http://www.zope.com/t1', "g"), x=u"vx", y=u"vy")
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Unknown directive', 'http://www.zope.com/t1', 'g')

        >>> context = ConfigurationMachine()
        >>> defineGroupingDirective(context, 'g', Ixy,
        ...                         GroupingContextDecorator, "*")

        >>> context.begin(('http://www.zope.com/t1', "g"), x=u"vx", y=u"vy")
        >>> context.stack[-1].context.x
        'vx'
        >>> context.stack[-1].context.y
        'vy'
    """  # noqa: E501 line too long
    namespace = namespace or context.namespace
    if namespace != '*':
        name = namespace, name

    def factory(context, data, info):
        args = toargs(context, schema, data)
        newcontext = handler(context, **args)
        newcontext.info = info
        return GroupingStackItem(newcontext)
    factory.schema = schema

    context.register(usedIn, name, factory)
    context.document(name, schema, usedIn, handler, context.info)


class IComplexDirectiveContext(IFullInfo, IConfigurationContext):
    pass


@implementer(IComplexDirectiveContext)
class ComplexDirectiveDefinition(GroupingContextDecorator, dict):
    """Handler for defining complex directives

    See the description and tests for ComplexStackItem.
    """

    def before(self):

        def factory(context, data, info):
            return ComplexStackItem(self, context, data, info)
        factory.schema = self.schema

        self.register(self.usedIn, (self.namespace, self.name), factory)
        self.document((self.namespace, self.name), self.schema, self.usedIn,
                      self.handler, self.info)


def subdirective(context, name, schema):
    context.document((context.namespace, name), schema, context.usedIn,
                     getattr(context.handler, name, context.handler),
                     context.info, context.context)
    context.context[name] = schema, context.info

##############################################################################
# Features


class IProvidesDirectiveInfo(Interface):
    """Information for a <meta:provides> directive"""

    feature = TextLine(
        title=u"Feature name",
        description=u"""The name of the feature being provided

        You can test available features with zcml:condition="have featurename".
        """,
    )


def provides(context, feature):
    """
    Declare that a feature is provided in context.

    Example:

        >>> from zope.configuration.config import ConfigurationContext
        >>> from zope.configuration.config import provides
        >>> c = ConfigurationContext()
        >>> provides(c, 'apidoc')
        >>> c.hasFeature('apidoc')
        True

    Spaces are not allowed in feature names (this is reserved for
    providing many features with a single directive in the future).

        >>> provides(c, 'apidoc onlinehelp')
        Traceback (most recent call last):
          ...
        ValueError: Only one feature name allowed

        >>> c.hasFeature('apidoc onlinehelp')
        False
    """
    if len(feature.split()) > 1:
        raise ValueError("Only one feature name allowed")
    context.provideFeature(feature)


##############################################################################
# Argument conversion

def toargs(context, schema, data):
    """
    Marshal data to an argument dictionary using a schema

    Names that are python keywords have an underscore added as a
    suffix in the schema and in the argument list, but are used
    without the underscore in the data.

    The fields in the schema must all implement IFromUnicode.

    All of the items in the data must have corresponding fields in the
    schema unless the schema has a true tagged value named
    'keyword_arguments'.

    Example:

        >>> from zope.configuration.config import toargs
        >>> from zope.schema import BytesLine
        >>> from zope.schema import Float
        >>> from zope.schema import Int
        >>> from zope.schema import TextLine
        >>> from zope.schema import URI
        >>> class schema(Interface):
        ...     in_ = Int(constraint=lambda v: v > 0)
        ...     f = Float()
        ...     n = TextLine(min_length=1, default=u"rob")
        ...     x = BytesLine(required=False)
        ...     u = URI()

        >>> context = ConfigurationMachine()
        >>> from pprint import PrettyPrinter
        >>> pprint = PrettyPrinter(width=50).pprint

        >>> pprint(toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2', 'n': u'bob', 'x': u'x.y.z',
        ...          'u': u'http://www.zope.org' }))
        {'f': 1.2,
         'in_': 1,
         'n': 'bob',
         'u': 'http://www.zope.org',
         'x': b'x.y.z'}

    If we have extra data, we'll get an error:

        >>> toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2', 'n': u'bob', 'x': u'x.y.z',
        ...          'u': u'http://www.zope.org', 'a': u'1'})
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Unrecognized parameters:', 'a')

    Unless we set a tagged value to say that extra arguments are ok:

        >>> schema.setTaggedValue('keyword_arguments', True)

        >>> pprint(toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2', 'n': u'bob', 'x': u'x.y.z',
        ...          'u': u'http://www.zope.org', 'a': u'1'}))
        {'a': '1',
         'f': 1.2,
         'in_': 1,
         'n': 'bob',
         'u': 'http://www.zope.org',
         'x': b'x.y.z'}

    If we omit required data we get an error telling us what was
    omitted:

        >>> pprint(toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2', 'n': u'bob', 'x': u'x.y.z'}))
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Missing parameter:', 'u')

    Although we can omit not-required data:

        >>> pprint(toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2', 'n': u'bob',
        ...          'u': u'http://www.zope.org', 'a': u'1'}))
        {'a': '1',
         'f': 1.2,
         'in_': 1,
         'n': 'bob',
         'u': 'http://www.zope.org'}

    And we can omit required fields if they have valid defaults
    (defaults that are valid values):

        >>> pprint(toargs(context, schema,
        ...        {'in': u'1', 'f': u'1.2',
        ...          'u': u'http://www.zope.org', 'a': u'1'}))
        {'a': '1',
         'f': 1.2,
         'in_': 1,
         'n': 'rob',
         'u': 'http://www.zope.org'}

    We also get an error if any data was invalid:

        >>> pprint(toargs(context, schema,
        ...        {'in': u'0', 'f': u'1.2', 'n': u'bob', 'x': u'x.y.z',
        ...          'u': u'http://www.zope.org', 'a': u'1'}))
        Traceback (most recent call last):
        ...
        ConfigurationError: ('Invalid value for', 'in', '0')
    """
    data = dict(data)
    args = {}
    for name, field in schema.namesAndDescriptions(True):
        field = field.bind(context)
        n = name
        if n.endswith('_') and iskeyword(n[:-1]):
            n = n[:-1]

        s = data.get(n, data)
        if s is not data:
            s = text_type(s)
            del data[n]

            try:
                args[str(name)] = field.fromUnicode(s)
            except ValidationError as v:
                reraise(
                    ConfigurationError(
                        "Invalid value for %r" % (n)).add_details(v),
                    None, sys.exc_info()[2])
        elif field.required:
            # if the default is valid, we can use that:
            default = field.default
            try:
                field.validate(default)
            except ValidationError as v:
                reraise(
                    ConfigurationError(
                        "Missing parameter: %r" % (n,)).add_details(v),
                    None, sys.exc_info()[2])
            args[str(name)] = default

    if data:
        # we had data left over
        try:
            keyword_arguments = schema.getTaggedValue('keyword_arguments')
        except KeyError:
            keyword_arguments = False
        if not keyword_arguments:
            raise ConfigurationError("Unrecognized parameters:", *data)

        for name in data:
            args[str(name)] = data[name]

    return args

##############################################################################
# Conflict resolution


def expand_action(discriminator, callable=None, args=(), kw=None,
                  includepath=(), info=None, order=0, **extra):
    if kw is None:
        kw = {}
    action = extra
    action.update(
        dict(
            discriminator=discriminator,
            callable=callable,
            args=args,
            kw=kw,
            includepath=includepath,
            info=info,
            order=order,
        )
    )
    return action


def resolveConflicts(actions):
    """
    Resolve conflicting actions.

    Given an actions list, identify and try to resolve conflicting actions.
    Actions conflict if they have the same non-None discriminator.
    Conflicting actions can be resolved if the include path of one of
    the actions is a prefix of the includepaths of the other
    conflicting actions and is unequal to the include paths in the
    other conflicting actions.
    """

    # organize actions by discriminators
    unique = {}
    output = []
    for i, action in enumerate(actions):
        if not isinstance(action, dict):
            # old-style tuple action
            action = expand_action(*action)

        # "order" is an integer grouping. Actions in a lower order will be
        # executed before actions in a higher order.  Within an order,
        # actions are executed sequentially based on original action ordering
        # ("i").
        order = action['order'] or 0
        discriminator = action['discriminator']

        # "ainfo" is a tuple of (order, i, action) where "order" is a
        # user-supplied grouping, "i" is an integer expressing the relative
        # position of this action in the action list being resolved, and
        # "action" is an action dictionary.  The purpose of an ainfo is to
        # associate an "order" and an "i" with a particular action; "order"
        # and "i" exist for sorting purposes after conflict resolution.
        ainfo = (order, i, action)

        if discriminator is None:
            # The discriminator is None, so this action can never conflict.
            # We can add it directly to the result.
            output.append(ainfo)
            continue

        L = unique.setdefault(discriminator, [])
        L.append(ainfo)

    # Check for conflicts
    conflicts = {}

    for discriminator, ainfos in unique.items():

        # We use (includepath, order, i) as a sort key because we need to
        # sort the actions by the paths so that the shortest path with a
        # given prefix comes first.  The "first" action is the one with the
        # shortest include path.  We break sorting ties using "order", then
        # "i".
        def bypath(ainfo):
            path, order, i = ainfo[2]['includepath'], ainfo[0], ainfo[1]
            return path, order, i

        ainfos.sort(key=bypath)
        ainfo, rest = ainfos[0], ainfos[1:]
        output.append(ainfo)
        _, _, action = ainfo
        basepath, baseinfo, discriminator = (action['includepath'],
                                             action['info'],
                                             action['discriminator'])

        for _, _, action in rest:
            includepath = action['includepath']
            # Test whether path is a prefix of opath
            if (includepath[:len(basepath)] != basepath  # not a prefix
                    or includepath == basepath):
                L = conflicts.setdefault(discriminator, [baseinfo])
                L.append(action['info'])

    if conflicts:
        raise ConfigurationConflictError(conflicts)

    # Sort conflict-resolved actions by (order, i) and return them.
    return [x[2] for x in sorted(output, key=operator.itemgetter(0, 1))]


class ConfigurationConflictError(ConfigurationError):

    def __init__(self, conflicts):
        super(ConfigurationConflictError, self).__init__()
        self._conflicts = conflicts

    def _with_details(self, opening, detail_formatter):
        r = ["Conflicting configuration actions"]
        for discriminator, infos in sorted(self._conflicts.items()):
            r.append("  For: %s" % (discriminator, ))
            for info in infos:
                for line in text_type(info).rstrip().split(u'\n'):
                    r.append(u"    " + line)

        opening = "\n".join(r)
        return super(ConfigurationConflictError, self)._with_details(
            opening, detail_formatter)


##############################################################################
# Bootstap code


def _bootstrap(context):

    # Set enough machinery to register other directives

    # Define the directive (simple directive) directive by calling it's
    # handler directly

    info = 'Manually registered in zope/configuration/config.py'

    context.info = info
    defineSimpleDirective(
        context,
        namespace=metans, name='directive',
        schema=IStandaloneDirectiveInfo,
        handler=defineSimpleDirective)
    context.info = ''

    # OK, now that we have that, we can use the machine to define the
    # other directives. This isn't the easiest way to proceed, but it lets
    # us eat our own dogfood. :)

    # Standalone groupingDirective
    context((metans, 'directive'),
            info,
            name='groupingDirective',
            namespace=metans,
            handler="zope.configuration.config.defineGroupingDirective",
            schema="zope.configuration.config.IStandaloneDirectiveInfo"
            )

    # Now we can use the grouping directive to define the directives directive
    context((metans, 'groupingDirective'),
            info,
            name='directives',
            namespace=metans,
            handler="zope.configuration.config.DirectivesHandler",
            schema="zope.configuration.config.IDirectivesInfo"
            )

    # directive and groupingDirective inside directives
    context((metans, 'directive'),
            info,
            name='directive',
            namespace=metans,
            usedIn="zope.configuration.config.IDirectivesContext",
            handler="zope.configuration.config.defineSimpleDirective",
            schema="zope.configuration.config.IFullInfo"
            )
    context((metans, 'directive'),
            info,
            name='groupingDirective',
            namespace=metans,
            usedIn="zope.configuration.config.IDirectivesContext",
            handler="zope.configuration.config.defineGroupingDirective",
            schema="zope.configuration.config.IFullInfo"
            )

    # Setup complex directive directive, both standalone, and in
    # directives directive
    context((metans, 'groupingDirective'),
            info,
            name='complexDirective',
            namespace=metans,
            handler="zope.configuration.config.ComplexDirectiveDefinition",
            schema="zope.configuration.config.IStandaloneDirectiveInfo"
            )
    context((metans, 'groupingDirective'),
            info,
            name='complexDirective',
            namespace=metans,
            usedIn="zope.configuration.config.IDirectivesContext",
            handler="zope.configuration.config.ComplexDirectiveDefinition",
            schema="zope.configuration.config.IFullInfo"
            )

    # Finally, setup subdirective directive
    context((metans, 'directive'),
            info,
            name='subdirective',
            namespace=metans,
            usedIn="zope.configuration.config.IComplexDirectiveContext",
            handler="zope.configuration.config.subdirective",
            schema="zope.configuration.config.IDirectiveInfo"
            )

    # meta:provides
    context((metans, 'directive'),
            info,
            name='provides',
            namespace=metans,
            handler="zope.configuration.config.provides",
            schema="zope.configuration.config.IProvidesDirectiveInfo"
            )



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Configuration-specific schema fields
"""
import os
import sys
import warnings

from zope.interface import implementer
from zope.schema import Bool as schema_Bool
from zope.schema import DottedName
from zope.schema import Field
from zope.schema import InterfaceField
from zope.schema import List
from zope.schema import PythonIdentifier as schema_PythonIdentifier
from zope.schema import Text
from zope.schema import ValidationError
from zope.schema.interfaces import IFromUnicode
from zope.schema.interfaces import InvalidValue

from zope.configuration.exceptions import ConfigurationError
from zope.configuration.interfaces import InvalidToken
from zope.configuration._compat import implementer_if_needed

__all__ = [
    'Bool',
    'GlobalObject',
    'GlobalInterface',
    'MessageID',
    'Path',
    'PythonIdentifier',
    'Tokens',
]


class PythonIdentifier(schema_PythonIdentifier):
    r"""
    This class is like `zope.schema.PythonIdentifier`.


    Let's look at an example:

      >>> from zope.configuration.fields import PythonIdentifier
      >>> class FauxContext(object):
      ...     pass
      >>> context = FauxContext()
      >>> field = PythonIdentifier().bind(context)

    Let's test the fromUnicode method:

      >>> field.fromUnicode(u'foo')
      'foo'
      >>> field.fromUnicode(u'foo3')
      'foo3'
      >>> field.fromUnicode(u'_foo3')
      '_foo3'

    Now let's see whether validation works alright

      >>> values = (u'foo', u'foo3', u'foo_', u'_foo3', u'foo_3', u'foo3_')
      >>> for value in values:
      ...     _ = field.fromUnicode(value)
      >>> from zope.schema import ValidationError
      >>> for value in (u'3foo', u'foo:', u'\\', u''):
      ...     try:
      ...         field.fromUnicode(value)
      ...     except ValidationError:
      ...         print('Validation Error ' + repr(value))
      Validation Error '3foo'
      Validation Error 'foo:'
      Validation Error '\\'
      Validation Error ''

    .. versionchanged:: 4.2.0
       Extend `zope.schema.PythonIdentifier`, which implies that `fromUnicode`
       validates the strings.
    """

    def _validate(self, value):
        super(PythonIdentifier, self)._validate(value)
        if not value:
            raise ValidationError(value).with_field_and_value(self, value)


@implementer_if_needed(IFromUnicode)
class GlobalObject(Field):
    """
    An object that can be accessed as a module global.

    The special value ``*`` indicates a value of `None`; this is
    not validated against the *value_type*.
    """

    _DOT_VALIDATOR = DottedName()

    def __init__(self, value_type=None, **kw):
        self.value_type = value_type
        super(GlobalObject, self).__init__(**kw)

    def _validate(self, value):
        super(GlobalObject, self)._validate(value)
        if self.value_type is not None:
            self.value_type.validate(value)

    def fromUnicode(self, value):
        r"""
        Find and return the module global at the path *value*.

          >>> d = {'x': 1, 'y': 42, 'z': 'zope'}
          >>> class fakeresolver(dict):
          ...     def resolve(self, n):
          ...         return self[n]
          >>> fake = fakeresolver(d)

          >>> from zope.schema import Int
          >>> from zope.configuration.fields import GlobalObject
          >>> g = GlobalObject(value_type=Int())
          >>> gg = g.bind(fake)
          >>> gg.fromUnicode("x")
          1
          >>> gg.fromUnicode("   x  \n  ")
          1
          >>> gg.fromUnicode("y")
          42
          >>> gg.fromUnicode("z")
          Traceback (most recent call last):
          ...
          WrongType: ('zope', (<type 'int'>, <type 'long'>), '')

          >>> g = GlobalObject(constraint=lambda x: x%2 == 0)
          >>> gg = g.bind(fake)
          >>> gg.fromUnicode("x")
          Traceback (most recent call last):
          ...
          ConstraintNotSatisfied: 1
          >>> gg.fromUnicode("y")
          42
          >>> g = GlobalObject()
          >>> gg = g.bind(fake)
          >>> print(gg.fromUnicode('*'))
          None
        """
        name = str(value.strip())

        # special case, mostly for interfaces
        if name == '*':
            return None

        try:
            # Leading dots are allowed here to indicate current
            # package, but not accepted by DottedName. Take care,
            # though, because a single dot is valid to resolve, but
            # not valid to pass to DottedName (as an empty string)
            to_validate = name.lstrip('.')
            if to_validate:
                self._DOT_VALIDATOR.validate(to_validate)
        except ValidationError as v:
            v.with_field_and_value(self, name)
            raise

        try:
            value = self.context.resolve(name)
        except ConfigurationError as v:
            raise ValidationError(v).with_field_and_value(self, name)

        self.validate(value)
        return value


@implementer_if_needed(IFromUnicode)
class GlobalInterface(GlobalObject):
    """
    An interface that can be accessed from a module.

    Example:

    First, we need to set up a stub name resolver:

      >>> from zope.interface import Interface
      >>> class IFoo(Interface):
      ...     pass
      >>> class Foo(object):
      ...     pass
      >>> d = {'Foo': Foo, 'IFoo': IFoo}
      >>> class fakeresolver(dict):
      ...     def resolve(self, n):
      ...         return self[n]
      >>> fake = fakeresolver(d)

    Now verify constraints are checked correctly:

      >>> from zope.configuration.fields import GlobalInterface
      >>> g = GlobalInterface()
      >>> gg = g.bind(fake)
      >>> gg.fromUnicode('IFoo') is IFoo
      True
      >>> gg.fromUnicode('  IFoo  ') is IFoo
      True
      >>> gg.fromUnicode('Foo')
      Traceback (most recent call last):
      ...
      NotAnInterface: (<class 'Foo'>, ...

    """

    def __init__(self, **kw):
        super(GlobalInterface, self).__init__(InterfaceField(), **kw)


@implementer(IFromUnicode)
class Tokens(List):
    """
    A list that can be read from a space-separated string.
    """

    def fromUnicode(self, value):
        r"""
        Split the input string and convert it to *value_type*.

        Consider GlobalObject tokens:

        First, we need to set up a stub name resolver:

          >>> d = {'x': 1, 'y': 42, 'z': 'zope', 'x.y.x': 'foo'}
          >>> class fakeresolver(dict):
          ...     def resolve(self, n):
          ...         return self[n]
          >>> fake = fakeresolver(d)

          >>> from zope.configuration.fields import Tokens
          >>> from zope.configuration.fields import GlobalObject
          >>> g = Tokens(value_type=GlobalObject())
          >>> gg = g.bind(fake)
          >>> gg.fromUnicode("  \n  x y z  \n")
          [1, 42, 'zope']

          >>> from zope.schema import Int
          >>> g = Tokens(value_type=
          ...            GlobalObject(value_type=
          ...                         Int(constraint=lambda x: x%2 == 0)))
          >>> gg = g.bind(fake)
          >>> gg.fromUnicode("x y")
          Traceback (most recent call last):
          ...
          InvalidToken: 1 in x y

          >>> gg.fromUnicode("z y")
          Traceback (most recent call last):
          ...
          InvalidToken: ('zope', (<type 'int'>, <type 'long'>), '') in z y
          >>> gg.fromUnicode("y y")
          [42, 42]
        """
        value = value.strip()
        if value:
            vt = self.value_type.bind(self.context)
            values = []
            for s in value.split():
                try:
                    v = vt.fromUnicode(s)
                except ValidationError as ex:
                    raise InvalidToken(
                        "%s in %r" % (ex, value)).with_field_and_value(
                            self, s)
                else:
                    values.append(v)
        else:
            values = []

        self.validate(values)

        return values


class PathProcessor(object):
    # Internal helper for manipulations on paths

    @classmethod
    def expand(cls, filename):
        # Perform the expansions we want to have done. Returns a
        # tuple: (path, needs_processing) If the second value is true,
        # further processing should be done (the path isn't fully
        # resolved); if false, the path should be used as is

        filename = filename.strip()
        # expanding a ~ at the front should generally result
        # in an absolute path.
        filename = os.path.expanduser(filename)
        filename = os.path.expandvars(filename)
        if os.path.isabs(filename):
            return os.path.normpath(filename), False
        return filename, True


@implementer_if_needed(IFromUnicode)
class Path(Text):
    """
    A file path name, which may be input as a relative path

    Input paths are converted to absolute paths and normalized.
    """

    def fromUnicode(self, value):
        r"""
        Convert the input path to a normalized, absolute path.

        Let's look at an example:

        First, we need a "context" for the field that has a path
        function for converting relative path to an absolute path.

        We'll be careful to do this in an operating system independent fashion.

          >>> from zope.configuration.fields import Path
          >>> class FauxContext(object):
          ...    def path(self, p):
          ...       return os.path.join(os.sep, 'faux', 'context', p)
          >>> context = FauxContext()
          >>> field = Path().bind(context)

        Lets try an absolute path first:

          >>> import os
          >>> p = os.path.join(os.sep, u'a', u'b')
          >>> n = field.fromUnicode(p)
          >>> n.split(os.sep)
          ['', 'a', 'b']

        This should also work with extra spaces around the path:

          >>> p = "   \n   %s   \n\n   " % p
          >>> n = field.fromUnicode(p)
          >>> n.split(os.sep)
          ['', 'a', 'b']

        Environment variables are expanded:

          >>> os.environ['path-test'] = '42'
          >>> with_env = os.path.join(os.sep, u'a', u'${path-test}')
          >>> n = field.fromUnicode(with_env)
          >>> n.split(os.sep)
          ['', 'a', '42']

        Now try a relative path:

          >>> p = os.path.join(u'a', u'b')
          >>> n = field.fromUnicode(p)
          >>> n.split(os.sep)
          ['', 'faux', 'context', 'a', 'b']

        The current user is expanded (these are implicitly relative paths):

          >>> old_home = os.environ.get('HOME')
          >>> os.environ['HOME'] = os.path.join(os.sep, 'HOME')
          >>> n = field.fromUnicode('~')
          >>> n.split(os.sep)
          ['', 'HOME']
          >>> if old_home:
          ...    os.environ['HOME'] = old_home
          ... else:
          ...    del os.environ['HOME']


        .. versionchanged:: 4.2.0
            Start expanding home directories and environment variables.
        """
        filename, needs_processing = PathProcessor.expand(value)
        if needs_processing:
            filename = self.context.path(filename)

        return filename


@implementer_if_needed(IFromUnicode)
class Bool(schema_Bool):
    """
    A boolean value.

    Values may be input (in upper or lower case) as any of:

    - yes / no
    - y / n
    - true / false
    - t / f

    .. caution::

       Do not confuse this with :class:`zope.schema.Bool`.
       That class will only parse ``"True"`` and ``"true"`` as
       `True` values. Any other value will silently be accepted as
       `False`. This class raises a validation error for unrecognized
       input.

    """

    def fromUnicode(self, value):
        """
        Convert the input string to a boolean.

        Example:

            >>> from zope.configuration.fields import Bool
            >>> Bool().fromUnicode(u"yes")
            True
            >>> Bool().fromUnicode(u"y")
            True
            >>> Bool().fromUnicode(u"true")
            True
            >>> Bool().fromUnicode(u"no")
            False
            >>> Bool().fromUnicode(u"surprise")
            Traceback (most recent call last):
            ...
            zope.schema._bootstrapinterfaces.InvalidValue
        """
        value = value.lower()
        if value in ('1', 'true', 'yes', 't', 'y'):
            return True
        if value in ('0', 'false', 'no', 'f', 'n'):
            return False
        # Unlike the superclass, anything else is invalid.
        raise InvalidValue().with_field_and_value(self, value)


@implementer_if_needed(IFromUnicode)
class MessageID(Text):
    """
    Text string that should be translated.

    When a string is converted to a message ID, it is also recorded in
    the context.
    """

    __factories = {}

    def fromUnicode(self, value):
        """
        Translate a string to a MessageID.

          >>> from zope.configuration.fields import MessageID
          >>> class Info(object):
          ...     file = 'file location'
          ...     line = 8
          >>> class FauxContext(object):
          ...     i18n_strings = {}
          ...     info = Info()
          >>> context = FauxContext()
          >>> field = MessageID().bind(context)

        There is a fallback domain when no domain has been specified.

        Exchange the warn function so we can make test whether the warning
        has been issued

          >>> warned = None
          >>> def fakewarn(*args, **kw):
          ...     global warned
          ...     warned = args

          >>> import warnings
          >>> realwarn = warnings.warn
          >>> warnings.warn = fakewarn

          >>> i = field.fromUnicode(u"Hello world!")
          >>> i
          'Hello world!'
          >>> i.domain
          'untranslated'
          >>> warned
          ("You did not specify an i18n translation domain for the '' field in file location",)

          >>> warnings.warn = realwarn

        With the domain specified:

          >>> context.i18n_strings = {}
          >>> context.i18n_domain = 'testing'

        We can get a message id:

          >>> i = field.fromUnicode(u"Hello world!")
          >>> i
          'Hello world!'
          >>> i.domain
          'testing'

        In addition, the string has been registered with the context:

          >>> context.i18n_strings
          {'testing': {'Hello world!': [('file location', 8)]}}

          >>> i = field.fromUnicode(u"Foo Bar")
          >>> i = field.fromUnicode(u"Hello world!")
          >>> from pprint import PrettyPrinter
          >>> pprint=PrettyPrinter(width=70).pprint
          >>> pprint(context.i18n_strings)
          {'testing': {'Foo Bar': [('file location', 8)],
                       'Hello world!': [('file location', 8),
                                        ('file location', 8)]}}

          >>> from zope.i18nmessageid import Message
          >>> isinstance(list(context.i18n_strings['testing'].keys())[0],
          ...            Message)
          True

        Explicit Message IDs

          >>> i = field.fromUnicode(u'[View-Permission] View')
          >>> i
          'View-Permission'
          >>> i.default
          'View'

          >>> i = field.fromUnicode(u'[] [Some] text')
          >>> i
          '[Some] text'
          >>> i.default is None
          True

        """  # noqa: E501 line too long
        context = self.context
        domain = getattr(context, 'i18n_domain', '')
        if not domain:
            domain = 'untranslated'
            warnings.warn(
                "You did not specify an i18n translation domain for the "
                "'%s' field in %s" % (self.getName(), context.info.file)
            )
        if not isinstance(domain, str):
            # IZopeConfigure specifies i18n_domain as a BytesLine, but that's
            # wrong on Python 3, where the filesystem uses str, and hence
            # zope.i18n registers ITranslationDomain utilities with str names.
            # If we keep bytes, we can't find those utilities.
            enc = sys.getfilesystemencoding() or sys.getdefaultencoding()
            domain = domain.decode(enc)

        v = super(MessageID, self).fromUnicode(value)

        # Check whether there is an explicit message is specified
        default = None
        if v.startswith('[]'):
            v = v[2:].lstrip()
        elif v.startswith('['):
            end = v.find(']')
            default = v[end + 2:]
            v = v[1:end]

        # Convert to a message id, importing the factory, if necessary
        factory = self.__factories.get(domain)
        if factory is None:
            import zope.i18nmessageid
            factory = zope.i18nmessageid.MessageFactory(domain)
            self.__factories[domain] = factory

        msgid = factory(v, default)

        # Record the string we got for the domain
        i18n_strings = context.i18n_strings
        strings = i18n_strings.get(domain)
        if strings is None:
            strings = i18n_strings[domain] = {}
        locations = strings.setdefault(msgid, [])
        locations.append((context.info.file, context.info.line))

        return msgid



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Tests for the documentation.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import os.path
import unittest
import doctest

import manuel.capture
import manuel.codeblock
import manuel.doctest
import manuel.ignore
import manuel.testing

from zope.testing import renormalizing

checker = renormalizing.RENormalizing([
    # Python 3 unicode removed the "u".
    (re.compile("u('.*?')"), r"\1"),
    (re.compile('u(".*?")'), r"\1"),
    # Python 3 bytes added the "b".
    (re.compile("b('.*?')"), r"\1"),
    (re.compile('b(".*?")'), r"\1"),
])

optionflags = (
    doctest.NORMALIZE_WHITESPACE
    | doctest.ELLIPSIS
    | doctest.IGNORE_EXCEPTION_DETAIL
)


def test_suite():
    # zope.testrunner
    suite = unittest.TestSuite()
    here = os.path.dirname(os.path.abspath(__file__))
    while not os.path.exists(os.path.join(here, 'setup.py')):
        prev, here = here, os.path.dirname(here)
        if here == prev:
            # Let's avoid infinite loops at root
            class SkippedDocTests(unittest.TestCase):  # pragma: no cover

                @unittest.skip('Could not find setup.py')
                def test_docs(self):
                    pass

            suite.addTest(
                unittest.makeSuite(SkippedDocTests))  # pragma: no cover
            return suite  # pragma: no cover

    docs = os.path.join(here, 'docs')
    api_docs = os.path.join(docs, 'api')

    doc_files_to_test = (
        'narr.rst',
    )

    api_files_to_test = (
        'config.rst',
        'docutils.rst',
        'fields.rst',
        'xmlconfig.rst',
    )

    # Plain doctest suites
    api_to_test = (
        'config',
        'docutils',
        'fields',
        'interfaces',
        'name',
        'xmlconfig',
        'zopeconfigure',
    )

    paths = [os.path.join(docs, f) for f in doc_files_to_test]
    paths += [os.path.join(api_docs, f) for f in api_files_to_test]

    m = manuel.ignore.Manuel()
    m += manuel.doctest.Manuel(checker=checker, optionflags=optionflags)
    m += manuel.codeblock.Manuel()
    m += manuel.capture.Manuel()

    suite.addTest(
        manuel.testing.TestSuite(
            m,
            *paths
        )
    )

    for mod_name in api_to_test:
        mod_name = 'zope.configuration.' + mod_name
        suite.addTest(
            doctest.DocTestSuite(
                mod_name,
                checker=checker,
                optionflags=optionflags
            )
        )

    return suite


def load_tests(loader, standard_tests, pattern):
    # Pure unittest protocol.
    return test_suite()



##############################################################################
#
# Copyright (c) 20!2 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test zope.configuration.fields.
"""
import unittest

# pylint:disable=protected-access


class _ConformsToIFromUnicode(object):

    def _getTargetClass(self):
        raise NotImplementedError

    def _makeOne(self, *args, **kw):
        raise NotImplementedError

    def test_class_conforms_to_IFromUnicode(self):
        from zope.interface.verify import verifyClass
        from zope.schema.interfaces import IFromUnicode
        verifyClass(IFromUnicode, self._getTargetClass())

    def test_instance_conforms_to_IFromUnicode(self):
        from zope.interface.verify import verifyObject
        from zope.schema.interfaces import IFromUnicode
        verifyObject(IFromUnicode, self._makeOne())


class GlobalObjectTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import GlobalObject
        return GlobalObject

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test__validate_wo_value_type(self):
        go = self._makeOne(value_type=None)
        for value in [0, 0.0, (), [], set(), frozenset(), u'', b'']:
            go._validate(value)  # noraise

    def test__validate_w_value_type(self):
        from zope.schema import Text
        from zope.schema.interfaces import WrongType
        go = self._makeOne(value_type=Text())
        go.validate(u'')
        for value in [0, 0.0, (), [], set(), frozenset(), b'']:
            with self.assertRaises(WrongType):
                go._validate(value)

    def test_fromUnicode_w_star_and_extra_ws(self):
        go = self._makeOne()
        self.assertEqual(go.fromUnicode(' * '), None)

    def test_fromUnicode_w_resolve_fails(self):
        from zope.schema import ValidationError
        from zope.configuration.config import ConfigurationError

        class Context(object):
            _resolved = None

            def resolve(self, name):
                self._resolved = name
                raise ConfigurationError()
        go = self._makeOne()
        context = Context()
        bound = go.bind(context)
        with self.assertRaises(ValidationError) as exc:
            bound.fromUnicode('tried')
        self.assertEqual(context._resolved, 'tried')
        ex = exc.exception
        self.assertIs(ex.field, bound)
        self.assertEqual(ex.value, 'tried')

    def test_fromUnicode_w_resolve_success(self):
        _target = object()

        class Context(object):
            _resolved = None

            def resolve(self, name):
                self._resolved = name
                return _target
        go = self._makeOne()
        context = Context()
        bound = go.bind(context)
        found = bound.fromUnicode('tried')
        self.assertIs(found, _target)
        self.assertEqual(context._resolved, 'tried')

    def test_fromUnicode_w_resolve_dots(self):
        _target = object()

        class Context(object):
            _resolved = None

            def resolve(self, name):
                self._resolved = name
                return _target
        go = self._makeOne()
        context = Context()
        bound = go.bind(context)
        for name in (
                '.',
                '..',
                '.foo',
                '..foo.bar'
        ):
            __traceback_info__ = name
            found = bound.fromUnicode(name)
            self.assertIs(found, _target)
            self.assertEqual(context._resolved, name)

    def test_fromUnicode_w_resolve_but_validation_fails(self):
        from zope.schema import Text
        from zope.schema import ValidationError
        _target = object()

        class Context(object):
            _resolved = None

            def resolve(self, name):
                self._resolved = name
                return _target
        go = self._makeOne(value_type=Text())
        context = Context()
        bound = go.bind(context)
        with self.assertRaises(ValidationError):
            bound.fromUnicode('tried')
        self.assertEqual(context._resolved, 'tried')

    def test_fromUnicode_rejects_slash(self):
        from zope.schema import ValidationError
        field = self._makeOne()
        with self.assertRaises(ValidationError) as exc:
            field.fromUnicode('foo/bar')
        ex = exc.exception
        self.assertIs(ex.field, field)
        self.assertEqual(ex.value, 'foo/bar')


class GlobalInterfaceTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import GlobalInterface
        return GlobalInterface

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_ctor(self):
        from zope.schema import InterfaceField
        gi = self._makeOne()
        self.assertIsInstance(gi.value_type, InterfaceField)


class TokensTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import Tokens
        return Tokens

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_fromUnicode_empty(self):
        tok = self._makeOne()
        self.assertEqual(tok.fromUnicode(''), [])

    def test_fromUnicode_strips_ws(self):
        from zope.schema import Text
        tok = self._makeOne(value_type=Text())
        self.assertEqual(tok.fromUnicode(u' one two three '),
                         [u'one', u'two', u'three'])

    def test_fromUnicode_invalid(self):
        from zope.schema import Int
        from zope.configuration.interfaces import InvalidToken
        tok = self._makeOne(value_type=Int(min=0))
        with self.assertRaises(InvalidToken) as exc:
            tok.fromUnicode(u' 1 -1 3 ')

        ex = exc.exception
        self.assertIs(ex.field, tok)
        self.assertEqual(ex.value, '-1')


class PathTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import Path
        return Path

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_fromUnicode_absolute(self):
        import os
        path = self._makeOne()
        self.assertEqual(path.fromUnicode('/'), os.path.normpath('/'))

    def test_fromUnicode_relative(self):
        class Context(object):
            _pathed = None

            def path(self, value):
                self._pathed = value
                return '/hard/coded'
        context = Context()
        path = self._makeOne()
        bound = path.bind(context)
        self.assertEqual(bound.fromUnicode('relative/path'), '/hard/coded')
        self.assertEqual(context._pathed, 'relative/path')


class BoolTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import Bool
        return Bool

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_fromUnicode_w_true_values(self):
        values = ['1', 'true', 'yes', 't', 'y']
        values += [x.upper() for x in values]
        bo = self._makeOne()
        for value in values:
            self.assertEqual(bo.fromUnicode(value), True)

    def test_fromUnicode_w_false_values(self):
        values = ['0', 'false', 'no', 'f', 'n']
        values += [x.upper() for x in values]
        bo = self._makeOne()
        for value in values:
            self.assertEqual(bo.fromUnicode(value), False)

    def test_fromUnicode_w_invalid(self):
        from zope.schema.interfaces import InvalidValue
        bo = self._makeOne()
        with self.assertRaises(InvalidValue) as exc:
            bo.fromUnicode('notvalid')
        ex = exc.exception
        self.assertIs(ex.field, bo)
        self.assertEqual(ex.value, 'notvalid')


class MessageIDTests(unittest.TestCase, _ConformsToIFromUnicode):

    def _getTargetClass(self):
        from zope.configuration.fields import MessageID
        return MessageID

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def _makeContext(self, domain='testing_domain'):
        class Info(object):
            file = 'test_file'
            line = 42

        class Context(object):
            i18n_domain = domain

            def __init__(self):
                self.i18n_strings = {}
                self.info = Info()
        return Context()

    def test_wo_domain(self):
        import warnings
        mid = self._makeOne()
        context = self._makeContext(None)
        bound = mid.bind(context)
        with warnings.catch_warnings(record=True) as log:
            msgid = bound.fromUnicode(u'testing')
        self.assertEqual(len(log), 1)
        self.assertTrue(str(log[0].message).startswith(
            'You did not specify an i18n translation domain'))
        self.assertEqual(msgid, 'testing')
        self.assertEqual(msgid.default, None)
        self.assertEqual(msgid.domain, 'untranslated')
        self.assertEqual(context.i18n_strings,
                         {'untranslated': {'testing': [('test_file', 42)]}})

    def test_w_empty_id(self):
        import warnings
        mid = self._makeOne()
        context = self._makeContext()
        bound = mid.bind(context)
        with warnings.catch_warnings(record=True) as log:
            msgid = bound.fromUnicode(u'[] testing')
        self.assertEqual(len(log), 0)
        self.assertEqual(msgid, 'testing')
        self.assertEqual(msgid.default, None)
        self.assertEqual(msgid.domain, 'testing_domain')
        self.assertEqual(context.i18n_strings,
                         {'testing_domain': {'testing': [('test_file', 42)]}})

    def test_w_id_and_default(self):
        import warnings
        mid = self._makeOne()
        context = self._makeContext()
        bound = mid.bind(context)
        with warnings.catch_warnings(record=True) as log:
            msgid = bound.fromUnicode(u'[testing] default')
        self.assertEqual(len(log), 0)
        self.assertEqual(msgid, 'testing')
        self.assertEqual(msgid.default, 'default')
        self.assertEqual(msgid.domain, 'testing_domain')
        self.assertEqual(context.i18n_strings,
                         {'testing_domain': {'testing': [('test_file', 42)]}})

    def test_domain_decodes_bytes(self):
        mid = self._makeOne()
        context = self._makeContext(domain=b'domain')
        bound = mid.bind(context)
        msgid = bound.fromUnicode(u'msgid')
        self.assertIsInstance(msgid.domain, str)
        self.assertEqual(msgid.domain, 'domain')
